/*---------------------------------------------------------------------------*/

/*!  \file  spring.c
     \brief  utility subroutines for coarse-grained slip-springs.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2010,2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <math.h>
#include "spring.h"

/*---------------------------------------------------------------------------*/
#define TOLERANCE 1.0e-12
#define MAX_TERMS_FOR_ASYMPTOTIC_EXPANSION 8
#define MAX_TERMS_FOR_TAYLOR_EXPANSION 64
#define NUM_EXPANSION_COEFFICIENTS 8

/*---------------------------------------------------------------------------*/
static double effective_slip_spring_strength_factor_for_small_n0(double n0);
static double effective_slip_spring_strength_factor_for_middle_n0(double n0);
static double effective_slip_spring_strength_factor_for_large_n0(double n0);

/*---------------------------------------------------------------------------*/
float effective_slip_spring_strength_factor(float n0)
{
    if(n0 < 0.1f)
    {
        return (float)effective_slip_spring_strength_factor_for_small_n0((double)n0);
    }
    else if(n0 < 10.0f)
    {
        return (float)effective_slip_spring_strength_factor_for_middle_n0((double)n0);
    }
    else
    {
        return (float)effective_slip_spring_strength_factor_for_large_n0((double)n0);
    }
}

/*---------------------------------------------------------------------------*/
double effective_slip_spring_strength_factor_for_small_n0(double n0)
{
    int i;
    double z,g,y;

    z = 1.0 / (exp(1.0 / n0) - 1.0);
    y = 0.0;
    g = n0 - z;

    for(i = 0;i < MAX_TERMS_FOR_ASYMPTOTIC_EXPANSION;i++)
    {
        y += g;

        g = (i + 1.0) * n0 * g - z / (i + 2.0);

        if((g / y) < TOLERANCE)
            break;
    }

    return y;
}

/*---------------------------------------------------------------------------*/
double effective_slip_spring_strength_factor_for_middle_n0(double n0)
{
    int i;
    double x,f,y;

    x = 1.0 / n0;
    y = 0.0;
    f = 1.0;

    for(i = 1;i < MAX_TERMS_FOR_TAYLOR_EXPANSION;i++)
    {
        f *= x / i;
        y += f / i;
        if((f / i / y) < TOLERANCE)
            break;
    }

    y /= exp(x) - 1.0;

    return y;
}

/*---------------------------------------------------------------------------*/
double effective_slip_spring_strength_factor_for_large_n0(double n0)
{
    int i;
    double x,y;
        /*  expansion coefficients  */
    static double a[NUM_EXPANSION_COEFFICIENTS] =
        {1.0e+0,
         -2.5e-1,
         1.3888888888888887e-2,
         3.472222222222222e-3,
         -3.0092592592592593e-4,
         -8.101851851851851e-5,
         7.401108591584782e-6,
         2.0175894431846814e-6};

    x = 1.0 / n0;
    y = 0.0;

    for(i = NUM_EXPANSION_COEFFICIENTS - 1;i >= 0;i--)
    {
        y = y * x + a[i];
    }

    return y;
}

/*---------------------------------------------------------------------------*/
