/*---------------------------------------------------------------------------*/

/*!  \file  chain.c
     \brief  definitions and subroutines for slip-springed chains.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2010,2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include "chain.h"
#include "spring.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
slip_springed_chain *allocate_slip_springed_chain(int m,int n)
{
    slip_springed_chain *chain;

    if((n < 2) || (m < 1))
        return NULL;

    chain = malloc(sizeof(slip_springed_chain));

    chain->n = n;
    chain->m = m;

    chain->r = malloc(sizeof(vector3) * n * m);
    chain->a = malloc(sizeof(vector3) * n * m);
    chain->phi = malloc(sizeof(int) * n * m);

    chain->f = malloc(sizeof(vector3) * n);
    chain->w = malloc(sizeof(vector3) * n);

    return chain;
}

/*---------------------------------------------------------------------------*/
void free_slip_springed_chain(slip_springed_chain *chain)
{
    free(chain->r);
    free(chain->a);
    free(chain->phi);

    free(chain->f);
    free(chain->w);

    free(chain);
}

/*---------------------------------------------------------------------------*/
void initialize_slip_springed_chain(slip_springed_chain *chain,rng_state *rng,
                                    double n0,double n_s,double zeta_s)
{
    int i,j;
    vector3 *r,*a;
    double psi0,q0;
    int *phi;

    chain->n0 = n0;
/*     chain->n_s_bar = n_s * effective_slip_spring_strength_factor(n0); */
    chain->n_s_bar = n_s;
    chain->zeta_s_inv = 1.0 / zeta_s;

    r = chain->r;
    a = chain->a;
    phi = chain->phi;

    psi0 = exp(-1.0 / chain->n0);
    q0 = sqrt(chain->n_s_bar / 3.0);

    for(j = 0;j < chain->m;j++)
    {
        r[0][0] = 0.0;
        r[0][1] = 0.0;
        r[0][2] = 0.0;

        for(i = 1;i < chain->n;i++)
        {
            r[i][0] = r[i - 1][0] + sqrt(1.0 / 3.0) * nrand(rng);
            r[i][1] = r[i - 1][1] + sqrt(1.0 / 3.0) * nrand(rng);
            r[i][2] = r[i - 1][2] + sqrt(1.0 / 3.0) * nrand(rng);
        }

        for(i = 0;i < chain->n;i++)
        {
            if(urand(rng) < psi0)
            {
                phi[i] = 0;
                a[i][0] = 0.0;
                a[i][1] = 0.0;
                a[i][2] = 0.0;
            }
            else
            {
                phi[i] = 1;
                a[i][0] = r[i][0] + q0 * nrand(rng);
                a[i][1] = r[i][1] + q0 * nrand(rng);
                a[i][2] = r[i][2] + q0 * nrand(rng);
            }
        }

        r += chain->n;
        a += chain->n;
        phi += chain->n;
    }
}

/*---------------------------------------------------------------------------*/
void calculate_average_stress_tensor(slip_springed_chain *chain,
                                     symmetric_tensor3 sigma)
{
    int i,j;
    vector3 *r;

    r = chain->r;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;

    for(j = 0;j < chain->m;j++)
    {
        for(i = 1;i < chain->n;i++)
        {
            sigma[0] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][0] - r[i - 1][0]);
            sigma[1] += 3 * (r[i][1] - r[i - 1][1]) * (r[i][1] - r[i - 1][1]);
            sigma[2] += 3 * (r[i][2] - r[i - 1][2]) * (r[i][2] - r[i - 1][2]);
            sigma[3] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][1] - r[i - 1][1]);
            sigma[4] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][2] - r[i - 1][2]);
            sigma[5] += 3 * (r[i][1] - r[i - 1][1]) * (r[i][2] - r[i - 1][2]);
        }

        r += chain->n;
    }

    sigma[0] /= chain->m;
    sigma[1] /= chain->m;
    sigma[2] /= chain->m;
    sigma[3] /= chain->m;
    sigma[4] /= chain->m;
    sigma[5] /= chain->m;
}

/*---------------------------------------------------------------------------*/
void calculate_average_virtual_stress_tensor(slip_springed_chain *chain,
                                             symmetric_tensor3 sigma)
{
    int i,j;
    vector3 *r,*a;
    int *phi;
    double k_s;

    r = chain->r;
    a = chain->a;
    phi = chain->phi;
    k_s = 3.0 / chain->n_s_bar;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;

    for(j = 0;j < chain->m;j++)
    {
        for(i = 0;i < chain->n;i++)
        {
            if(phi[i] == 1)
            {
                sigma[0] += k_s * (r[i][0] - a[i][0]) * (r[i][0] - a[i][0]);
                sigma[1] += k_s * (r[i][1] - a[i][1]) * (r[i][1] - a[i][1]);
                sigma[2] += k_s * (r[i][2] - a[i][2]) * (r[i][2] - a[i][2]);
                sigma[3] += k_s * (r[i][0] - a[i][0]) * (r[i][1] - a[i][1]);
                sigma[4] += k_s * (r[i][0] - a[i][0]) * (r[i][2] - a[i][2]);
                sigma[5] += k_s * (r[i][1] - a[i][1]) * (r[i][2] - a[i][2]);
            }
        }

        r += chain->n;
        a += chain->n;
        phi += chain->n;
    }

    sigma[0] /= chain->m;
    sigma[1] /= chain->m;
    sigma[2] /= chain->m;
    sigma[3] /= chain->m;
    sigma[4] /= chain->m;
    sigma[5] /= chain->m;
}

/*---------------------------------------------------------------------------*/
