/*---------------------------------------------------------------------------*/

/*!  \file  kernel_gpu.cu
     \brief  kernel functions of an entangled single chain simulator on GPU  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_KERNEL_GPU_CU
#define INCLUDE_KERNEL_GPU_CU

/*---------------------------------------------------------------------------*/
#include "threads_gpu.h"
#include "rng_gpu.cu"

/*---------------------------------------------------------------------------*/
__device__ __constant__ int m,n,z_max;
__device__ __constant__ float dt,zeta_s_inv,n_s,n0;
__device__ __constant__ float kappa[3][3];
__device__ __constant__ int iterations;

#if (__CUDA_ARCH__ < 120)
    /*  alignment size: 8 * 4 bytes = 32 bytes  */
__device__ __attribute__((aligned(32))) int f_global[N_BLOCKS * N_THREADS][8];
#endif

/*---------------------------------------------------------------------------*/
    /*  fixed real number:
          EPSILON = 1 / FLOAT2FIXED_CONVERSION_FACTOR
          MAX = INT_MAX / FLOAT2FIXED_CONVERSION_FACTOR
          MIN = 1 / FLOAT2FIXED_CONVERSION_FACTOR  */
#define FLOAT2FIXED_CONVERSION_FACTOR 0x00001000

#define SQR(x) ((x) * (x))

/*---------------------------------------------------------------------------*/
__device__ inline int float2fixed(float x)
{
    return ((int)(x * FLOAT2FIXED_CONVERSION_FACTOR));
}

/*---------------------------------------------------------------------------*/
__device__ inline float fixed2float(int x)
{
    return ((float)x / FLOAT2FIXED_CONVERSION_FACTOR);
}

/*---------------------------------------------------------------------------*/
__device__ inline void move_slip_spring(int *s,float w_plus,
                                        float w_minus,unsigned int *rng)
{
    float u;

    u = urand_gpu(rng);
    if(u < w_plus)
    {
        (*s)++;
    }
    else if(u < (w_plus + w_minus))
    {
        (*s)--;
    }
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_number_of_slip_springs(int i,int s,int *z_sub)
#if ENABLE_FAST_REDUCTION
{
        /*  this is a faster version of the reduction subroutine.
            (unfortunately, this does not work on some environments).
            define ENABLE_FAST_REDUCTION to use the faster version.  */

    z_sub[i] = ((s == NO_BEAD_INDEX) ? (0) : (1));
    __syncthreads();

        /*  reduction: z[0] + z[1] + ... + z[n - 1] -> z[0]  */

#if (N_THREADS >= 512)
    if(n >= 512)
    {
        if(i < 256)
            z_sub[i] += z_sub[i + 256];
        __syncthreads();
    }
#endif
#if (N_THREADS >= 256)
    if(n >= 256)
    {
        if(i < 128)
            z_sub[i] += z_sub[i + 128];
        __syncthreads();
    }
#endif
#if (N_THREADS >= 128)
    if(n >= 128)
    {
        if(i < 64)
            z_sub[i] += z_sub[i + 64];
        __syncthreads();
    }
#endif

    if(i < 32)
    {
        if(n >= 64)
            z_sub[i] += z_sub[i + 32];
        if(n >= 32)
            z_sub[i] += z_sub[i + 16];
        if(n >= 16)
            z_sub[i] += z_sub[i + 8];
        if(n >= 8)
            z_sub[i] += z_sub[i + 4];
        if(n >= 4)
            z_sub[i] += z_sub[i + 2];
        z_sub[i] += z_sub[i + 1];
    }
    __syncthreads();
}
#else
{
    int j;

    z_sub[i] = ((s == NO_BEAD_INDEX) ? (0) : (1));
    __syncthreads();

        /*  reduction: z[0] + z[1] + ... + z[n - 1] -> z[0]  */
    for(j = n / 2;j > 0;j >>= 1)
    {
        if(i < j)
        {
            z_sub[i] += z_sub[i + j];
        }
        __syncthreads();
     }
}
#endif

/*---------------------------------------------------------------------------*/
__device__ inline float slip_spring_potential(float *r,float *a)
{
    return (1.5f / n_s *
            (SQR(r[0] - a[0]) + SQR(r[1] - a[1]) + SQR(r[2] - a[2])));
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_chain_spring_force(float *r0,float *r1,
                                                    float *r2,float *f)
{
    f[0] = 3.0f * (r1[0] + r2[0] - 2.0f * r0[0]);
    f[1] = 3.0f * (r1[1] + r2[1] - 2.0f * r0[1]);
    f[2] = 3.0f * (r1[2] + r2[2] - 2.0f * r0[2]);
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_jump_probabilities(float (*r_sub)[3],
                                                    float *a,int s,
                                                    float *w_plus,
                                                    float *w_minus)
{
    float f_old,f_new;

    f_old = slip_spring_potential(r_sub[s],a);

    if(s < (n - 1))
    {
        f_new = slip_spring_potential(r_sub[s + 1],a);
        (*w_plus) = (2.0f * dt * zeta_s_inv) / (1.0f + expf(f_new - f_old));
    }
    else
    {
        (*w_plus) = 0.0f;
    }

    if(s > 0)
    {
        f_new = slip_spring_potential(r_sub[s - 1],a);
        (*w_minus) = (2.0f * dt * zeta_s_inv) / (1.0f + expf(f_new - f_old));
    }
    else
    {
        (*w_minus) = 0.0f;
    }
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_new_slip_spring_position(float *r,float *w,
                                                          float *a)
{
    a[0] = r[0] + sqrtf(n_s / 3.0f) * w[0];
    a[1] = r[1] + sqrtf(n_s / 3.0f) * w[1];
    a[2] = r[2] + sqrtf(n_s / 3.0f) * w[3];
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_stress_tensor(float *r0,float *r1,
                                               float *sigma)
{
    float u[3];

    u[0] = r1[0] - r0[0];
    u[1] = r1[1] - r0[1];
    u[2] = r1[2] - r0[2];

    sigma[0] = 3.0f * u[0] * u[0];
    sigma[1] = 3.0f * u[1] * u[1];
    sigma[2] = 3.0f * u[2] * u[2];
    sigma[3] = 3.0f * u[0] * u[1];
    sigma[4] = 3.0f * u[0] * u[2];
    sigma[5] = 3.0f * u[1] * u[2];
}

/*---------------------------------------------------------------------------*/
__device__ inline void add_float6(float *x,float *y)
{
    x[0] += y[0];
    x[1] += y[1];
    x[2] += y[2];
    x[3] += y[3];
    x[4] += y[4];
    x[5] += y[5];
}

/*---------------------------------------------------------------------------*/
__device__ inline void calculate_center_of_mass(int i,float *r,
                                                float (*rc_sub)[3])
{
    int j;

    rc_sub[i][0] = r[0] / n;
    rc_sub[i][1] = r[1] / n;
    rc_sub[i][2] = r[2] / n;
    __syncthreads();

        /*  reduction: rc[0] + rc[1] + ... + rc[n - 1] -> rc[0]  */

    for(j = n / 2;j > 0;j >>= 1)
    {
        if(i < j)
        {
            rc_sub[i][0] += rc_sub[i + j][0];
            rc_sub[i][1] += rc_sub[i + j][1];
            rc_sub[i][2] += rc_sub[i + j][2];
        }
        __syncthreads();
    }
}

/*---------------------------------------------------------------------------*/
#include "simulator_gpu.cu"

/*---------------------------------------------------------------------------*/
__global__ void calculate_subtotal_stress_tensor_gpu(float *r_device,
                                                     float *sigma_device)
{
    const int thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const int local_id = threadIdx.x;
    int index,block_index;
    int i,local_id_plus,j;
    __shared__ float r[N_THREADS][3];
    __shared__ float sigma[N_THREADS][6];

    block_index = blockIdx.x;
    i = local_id;
    while(i >= n)
    {
        i -= n;
    }
    local_id_plus = ((i == (n - 1)) ? (local_id) : (local_id + 1));

    for(index = thread_id;index < m * n;index += N_BLOCKS * N_THREADS)
    {
            /*  load data from global memory  */

        r[local_id][0] = r_device[index * 3];
        r[local_id][1] = r_device[index * 3 + 1];
        r[local_id][2] = r_device[index * 3 + 2];
        __syncthreads();


            /*  calculate stress tensors  */

        calculate_stress_tensor(r[local_id],r[local_id_plus],
                                sigma[local_id]);
        __syncthreads();


            /*  reduction: sigma[0] + sigma[1] + ... + sigma[N_THREADS]
                -> sigma[0]  */

        for(j = N_THREADS / 2;j > 0;j >>= 1)
        {
            if(local_id < j)
            {
                add_float6(sigma[local_id],sigma[local_id + j]);
            }
            __syncthreads();
        }


            /*  store data to global memory  */

        if(local_id == 0)
        {
            sigma_device[block_index * 6] = sigma[0][0];
            sigma_device[block_index * 6 + 1] = sigma[0][1];
            sigma_device[block_index * 6 + 2] = sigma[0][2];
            sigma_device[block_index * 6 + 3] = sigma[0][3];
            sigma_device[block_index * 6 + 4] = sigma[0][4];
            sigma_device[block_index * 6 + 5] = sigma[0][5];
        }
        __syncthreads();

        block_index += N_BLOCKS;
    }
}

/*---------------------------------------------------------------------------*/
__global__ void reduce_subtotal_stress_tensor_gpu(float *sigma_device,
                                                  float *sigma_reduced_device,
                                                  int n_items)
{
    const int thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const int local_id = threadIdx.x;
    int index,block_index;
    int j;
    __shared__ float sigma[N_THREADS][6];

    block_index = blockIdx.x;

    for(index = thread_id;index < n_items;index += N_BLOCKS * N_THREADS)
    {
            /*  load data from global memory  */

        sigma[local_id][0] = sigma_device[index * 6];
        sigma[local_id][1] = sigma_device[index * 6 + 1];
        sigma[local_id][2] = sigma_device[index * 6 + 2];
        sigma[local_id][3] = sigma_device[index * 6 + 3];
        sigma[local_id][4] = sigma_device[index * 6 + 4];
        sigma[local_id][5] = sigma_device[index * 6 + 5];
        __syncthreads();


            /*  reduction: sigma[0] + sigma[1] + ... + sigma[N_THREADS]
                -> sigma[0]  */

        for(j = N_THREADS / 2;j > 0;j >>= 1)
        {
            if((local_id < j) && (local_id + j < n_items))
            {
                add_float6(sigma[local_id],sigma[local_id + j]);
            }
            __syncthreads();
        }


            /*  store data to global memory  */

        if(local_id == 0)
        {
            sigma_reduced_device[block_index * 6] = sigma[0][0];
            sigma_reduced_device[block_index * 6 + 1] = sigma[0][1];
            sigma_reduced_device[block_index * 6 + 2] = sigma[0][2];
            sigma_reduced_device[block_index * 6 + 3] = sigma[0][3];
            sigma_reduced_device[block_index * 6 + 4] = sigma[0][4];
            sigma_reduced_device[block_index * 6 + 5] = sigma[0][5];
        }
        __syncthreads();

        block_index += N_BLOCKS;
    }
}

/*---------------------------------------------------------------------------*/
__global__ void set_centers_of_mass_to_origin_gpu(float *r_device,
                                                  float *a_device)
{
    const int thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const int local_id = threadIdx.x;
    int index;
    int i;
    float a[3],r[3];
    __shared__ float rc[N_THREADS][3];
    float (*rc_sub)[3];

    rc_sub = rc;
    i = local_id;
    while(i >= n)
    {
        rc_sub += n;
        i -= n;
    }

    for(index = thread_id;index < m * n;index += N_BLOCKS * N_THREADS)
    {
            /*  load data from global memory  */

        r[0] = r_device[index * 3];
        r[1] = r_device[index * 3 + 1];
        r[2] = r_device[index * 3 + 2];
        a[0] = a_device[index * 3];
        a[1] = a_device[index * 3 + 1];
        a[2] = a_device[index * 3 + 2];
        __syncthreads();


            /*  subtract center of mass position from bead and
                slip-spring positions  */

        calculate_center_of_mass(i,r,rc_sub);

        r[0] -= rc_sub[0][0];
        r[1] -= rc_sub[0][1];
        r[2] -= rc_sub[0][2];
        a[0] -= rc_sub[0][0];
        a[1] -= rc_sub[0][1];
        a[2] -= rc_sub[0][2];


            /*  store data to global memory  */

        r_device[index * 3] = r[0];
        r_device[index * 3 + 1] = r[1];
        r_device[index * 3 + 2] = r[2];
        a_device[index * 3] = a[0];
        a_device[index * 3 + 1] = a[1];
        a_device[index * 3 + 2] = a[2];
        __syncthreads();
    }
}

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
