/*---------------------------------------------------------------------------*/

/*!  \file  load.c
     \brief  load the state from a HDF5 file  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include "load.h"
#include <stdio.h>
#include <stdlib.h>
#include <hdf5.h>
#include <hdf5_hl.h>

/*---------------------------------------------------------------------------*/
static void load_rng_from_hdf5(hid_t file,simulator_state *state);
static void load_rng_gpu_from_hdf5(hid_t file,simulator_state *state);
static void load_polymer_chains_from_hdf5(hid_t file,simulator_state *state);
static void load_slip_springs_from_hdf5(hid_t file,simulator_state *state);
static void read_dataset_unsignedlong(hid_t location,char *path,
                                      unsigned long *dest);
static void read_dataset_unsignedint(hid_t location,char *path,
                                     unsigned int *dest);

/*---------------------------------------------------------------------------*/
simulator_state *load_state_from_file(char *h5filename)
{
    hid_t file;
    hid_t group;
    int m,n,z_max;
    simulator_state *state;

    file = H5Fopen(h5filename,H5F_ACC_RDONLY,H5P_DEFAULT);
    if(file < 0)
    {
        fprintf(stderr,"error: cannot open the input HDF5 file `%s'.\n",
                h5filename);
        exit(EXIT_FAILURE);
    }

    H5LTread_dataset_int(file,"/m",&m);

    group = H5Gopen2(file,"/polymer_chain",H5P_DEFAULT);
    H5LTread_dataset_int(file,"/polymer_chain/n",&n);
    H5Gclose(group);

    group = H5Gopen2(file,"/slip_springs",H5P_DEFAULT);
    H5LTread_dataset_int(file,"/slip_springs/z_max",&z_max);
    H5Gclose(group);

    state = allocate_simulator_state(m,n,z_max);

    load_rng_from_hdf5(file,state);
    load_rng_gpu_from_hdf5(file,state);
    load_polymer_chains_from_hdf5(file,state);
    load_slip_springs_from_hdf5(file,state);

    H5Fclose(file);

    return state;
}

/*---------------------------------------------------------------------------*/
void load_rng_from_hdf5(hid_t file,simulator_state *state)
{
    hid_t group1,group2;
    int next_offset;
    rng_state *rng;

    rng = state->rng;

    group1 = H5Gopen2(file,"/rng_state",H5P_DEFAULT);
    group2 = H5Gopen2(file,"/rng_state/mt_state",H5P_DEFAULT);

    read_dataset_unsignedlong(file,"/rng_state/mt_state/state",rng->mt->state);

    H5LTread_dataset_int(file,"/rng_state/mt_state/n",&(rng->mt->n));
    H5LTread_dataset_int(file,"/rng_state/mt_state/left",&(rng->mt->left));
    H5LTread_dataset_int(file,"/rng_state/mt_state/initf",&(rng->mt->initf));
    H5LTread_dataset_int(file,"/rng_state/mt_state/next",&next_offset);
    rng->mt->next = rng->mt->state + next_offset;

    H5LTread_dataset_int(file,"/rng_state/nrand_has_saved_number",
                         &(rng->nrand_has_saved_number));
    H5LTread_dataset_double(file,"/rng_state/nrand_w",&(rng->nrand_w));

    H5Gclose(group2);
    H5Gclose(group1);
}

/*---------------------------------------------------------------------------*/
void load_rng_gpu_from_hdf5(hid_t file,simulator_state *state)
{
    hid_t group;
    rng_state_gpu *rng_gpu;

    rng_gpu = state->rng_gpu;

    group = H5Gopen2(file,"/rng_gpu_state",H5P_DEFAULT);

    H5LTread_dataset_int(file,"/rng_gpu_state/n",&(rng_gpu->n));

    read_dataset_unsignedint(file,"/rng_gpu_state/state",
                             (rng_gpu->state_host));

    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void load_polymer_chains_from_hdf5(hid_t file,simulator_state *state)
{
    int i,n;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    group = H5Gopen2(file,"/polymer_chain",H5P_DEFAULT);

    n = state->n;

    data = H5Dopen2(file,"/polymer_chain/r",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = n;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = n;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < state->m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_FLOAT,slab,space,
                H5P_DEFAULT,(state->chains->r_host + 3 * n * i));
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);

    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void load_slip_springs_from_hdf5(hid_t file,simulator_state *state)
{
    int i,z_max;
    hid_t group,data,space,slab;
    hsize_t dims[3],start[3],stride[3],count[3],block[3];

    group = H5Gopen2(file,"/slip_springs",H5P_DEFAULT);

    z_max = state->z_max;
    H5LTread_dataset_float(file,"/slip_springs/n0",&(state->n0));
    H5LTread_dataset_float(file,"/slip_springs/n_s",&(state->n_s));
    H5LTread_dataset_float(file,"/slip_springs/zeta_s_inv",
                           &(state->zeta_s_inv));


    data = H5Dopen2(file,"/slip_springs/a",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = z_max;
    dims[1] = 3;
    slab = H5Screate_simple(2,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    start[2] = 0;
    stride[0] = 1;
    stride[1] = 1;
    stride[2] = 1;
    count[0] = 1;
    count[1] = z_max;
    count[2] = 3;
    block[0] = 1;
    block[1] = 1;
    block[2] = 1;
    for(i = 0;i < state->m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_FLOAT,slab,space,
                H5P_DEFAULT,(state->springs->a_host + 3 * z_max * i));
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);


    data = H5Dopen2(file,"/slip_springs/s",H5P_DEFAULT);
    space = H5Dget_space(data);
    dims[0] = z_max;
    slab = H5Screate_simple(1,dims,NULL);

    start[0] = 0;
    start[1] = 0;
    stride[0] = 1;
    stride[1] = 1;
    count[0] = 1;
    count[1] = z_max;
    block[0] = 1;
    block[1] = 1;
    for(i = 0;i < state->m;i++)
    {
        start[0] = i;
        H5Sselect_hyperslab(space,H5S_SELECT_SET,
                            start,stride,count,block);

        H5Dread(data,H5T_NATIVE_INT,slab,space,
                H5P_DEFAULT,(state->springs->s_host + z_max * i));
    }

    H5Sclose(slab);
    H5Sclose(space);

    H5Dclose(data);


    H5Gclose(group);
}

/*---------------------------------------------------------------------------*/
void read_dataset_unsignedlong(hid_t location,char *path,unsigned long *dest)
{
    hid_t data;

    data = H5Dopen2(location,path,H5P_DEFAULT);

    H5Dread(data,H5T_NATIVE_ULONG,H5S_ALL,H5S_ALL,H5P_DEFAULT,dest);

    H5Dclose(data);
}

/*---------------------------------------------------------------------------*/
void read_dataset_unsignedint(hid_t location,char *path,unsigned int *dest)
{
    hid_t data;

    data = H5Dopen2(location,path,H5P_DEFAULT);

    H5Dread(data,H5T_NATIVE_UINT,H5S_ALL,H5S_ALL,H5P_DEFAULT,dest);

    H5Dclose(data);
}

/*---------------------------------------------------------------------------*/
