/*---------------------------------------------------------------------------*/

/*!  \file  simulator_gpu.cu
     \brief  kernels for the simulator on GPU  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009-2011 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#ifndef INCLUDE_SIMULATOR_GPU_CU
#define INCLUDE_SIMULATOR_GPU_CU

/*---------------------------------------------------------------------------*/
__device__ inline void integrate_langevin_equation(float *r,float *a,
                                                   float *f,float *w)
{
    f[0] += kappa[0][0] * r[0] + kappa[0][1] * r[1] + kappa[0][2] * r[2];
    f[1] += kappa[1][0] * r[0] + kappa[1][1] * r[1] + kappa[1][2] * r[2];
    f[2] += kappa[2][0] * r[0] + kappa[2][1] * r[1] + kappa[2][2] * r[2];
    r[0] += dt * f[0] + sqrtf(2.0f * dt) * w[0];
    r[1] += dt * f[1] + sqrtf(2.0f * dt) * w[1];
    r[2] += dt * f[2] + sqrtf(2.0f * dt) * w[2];

     __syncthreads();

    f[0] = kappa[0][0] * a[0] + kappa[0][1] * a[1] + kappa[0][2] * a[2];
    f[1] = kappa[1][0] * a[0] + kappa[1][1] * a[1] + kappa[1][2] * a[2];
    f[2] = kappa[2][0] * a[0] + kappa[2][1] * a[1] + kappa[2][2] * a[2];
    a[0] += dt * f[0];
    a[1] += dt * f[1];
    a[2] += dt * f[2];
}

/*---------------------------------------------------------------------------*/
__device__ inline void reconstruct_slip_spring(int *s,float *a,
                                               float (*r_sub)[3],float *w,
                                               float w_construct,
                                               float w_destruct,
                                               unsigned int *rng)
{
    float u;

    u = urand_gpu(rng);

    if((*s) == NO_BEAD_INDEX)
    {
        u -= w_construct;

        if(u < 0)
        {
            (*s) = 0;

            nrand2_gpu(rng,&(w[0]),&(w[1]));
            calculate_new_slip_spring_position(r_sub[0],w,a);
        }
        else if(u < w_construct)
        {
            (*s) = n - 1;

            nrand2_gpu(rng,&(w[0]),&(w[1]));
            calculate_new_slip_spring_position(r_sub[n - 1],w,a);
        }
    }
    else
    {
        if(u < w_destruct)
        {
            (*s) = NO_BEAD_INDEX;
            a[0] = 0.0f;
            a[1] = 0.0f;
            a[2] = 0.0f;
        }
    }
}

/*---------------------------------------------------------------------------*/
__global__ void perform_simulation_gpu(float *r_device,float *a_device,
                                       int *s_device,unsigned int *rng_state)
{
    const int thread_id = blockDim.x * blockIdx.x + threadIdx.x;
    const int local_id = threadIdx.x;
    int index;
    int i,i_plus,i_minus,k;
    int s;
    float w[4];
    float a[3],f[3];
    float w_plus,w_minus;
    float w_construct,w_destruct;
    unsigned int rng;
    __shared__ int z[N_THREADS];
    __shared__ float r[N_THREADS][3];
    float (*r_sub)[3];
    int *z_sub;
#if (__CUDA_ARCH__ < 120)
    int (*f_fixed_sub)[8];
        /*  use global memory for atomicAdd() (slow but portable)  */
    int (*f_fixed)[8];

    f_fixed = f_global + N_THREADS * blockIdx.x;
#else
        /*  use shared memory for atomicAdd() (fast)  */
    int (*f_fixed_sub)[3];
    __shared__ int f_fixed[N_THREADS][3];
#endif


    load_rand_gpu(thread_id,rng_state,&rng);

    r_sub = r;
    f_fixed_sub = f_fixed;
    z_sub = z;
    i = local_id;
    while(i >= n)
    {
        r_sub += n;
        f_fixed_sub += n;
        z_sub += n;
        i -= n;
    }
    i_plus = ((i == (n - 1)) ? (n - 1) : (i + 1));
    i_minus = ((i == 0) ? (0) : (i - 1));


    for(index = thread_id;index < m * n;index += N_BLOCKS * N_THREADS)
    {
            /*  load data from global memory  */

        r[local_id][0] = r_device[index * 3];
        r[local_id][1] = r_device[index * 3 + 1];
        r[local_id][2] = r_device[index * 3 + 2];
        a[0] = a_device[index * 3];
        a[1] = a_device[index * 3 + 1];
        a[2] = a_device[index * 3 + 2];
        s = s_device[index];
        __syncthreads();



            /*  main iteration loop  */

        for(k = 0;k < iterations;k++)
        {
                /*  calculation of force  */
            
            calculate_chain_spring_force(r_sub[i],r_sub[i_minus],
                                         r_sub[i_plus],f);


            f_fixed[local_id][0] = 0;
            f_fixed[local_id][1] = 0;
            f_fixed[local_id][2] = 0;
            __syncthreads();

            if(s != NO_BEAD_INDEX)
            {
                atomicAdd(&(f_fixed_sub[s][0]),
                          float2fixed(3.0f / n_s * (a[0] - r_sub[s][0])));
                atomicAdd(&(f_fixed_sub[s][1]),
                          float2fixed(3.0f / n_s * (a[1] - r_sub[s][1])));
                atomicAdd(&(f_fixed_sub[s][2]),
                          float2fixed(3.0f / n_s * (a[2] - r_sub[s][2])));
            }
            __syncthreads();

            f[0] += fixed2float(f_fixed[local_id][0]);
            f[1] += fixed2float(f_fixed[local_id][1]);
            f[2] += fixed2float(f_fixed[local_id][2]);


            nrand2_gpu(&rng,&(w[0]),&(w[1]));
            nrand2_gpu(&rng,&(w[2]),&(w[3]));


                /*  integration of the Langevin equation  */
            integrate_langevin_equation(r[local_id],a,f,w);



            if(s != NO_BEAD_INDEX)
            {
                    /*  calculation of jump probabilities  */

                calculate_jump_probabilities(r_sub,a,s,&w_plus,&w_minus);


                    /*  move slip spring positions by jumps  */

                move_slip_spring(&s,w_plus,w_minus,&rng);
            }



                /*  calculation of reconstruction probabilities  */

            calculate_number_of_slip_springs(i,s,z_sub);

            if((s == 0) || (s == (n - 1)))
            {
                w_destruct = dt * zeta_s_inv;
            }
            else
            {
                w_destruct = 0.0f;
            }

            w_construct = ((z_sub[0] == z_max) ? (0.0f) :
                           (dt * zeta_s_inv / n0 / (z_max - z_sub[0])));


                /*  reconstruct slip-springs  */

            reconstruct_slip_spring(&s,a,r_sub,w,w_construct,w_destruct,&rng);
        }


            /*  store data to global memory  */

        r_device[index * 3] = r[local_id][0];
        r_device[index * 3 + 1] = r[local_id][1];
        r_device[index * 3 + 2] = r[local_id][2];
        a_device[index * 3] = a[0];
        a_device[index * 3 + 1] = a[1];
        a_device[index * 3 + 2] = a[2];
        s_device[index] = s;
        __syncthreads();
    }


    save_rand_gpu(thread_id,rng_state,&rng);
}

/*---------------------------------------------------------------------------*/
#endif

/*---------------------------------------------------------------------------*/
