/*---------------------------------------------------------------------------*/

/*!  \file  chain.c
     \brief  definitions and subroutines for a linear polymer chain.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <math.h>
#include "chain.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
polymer_chain *allocate_polymer_chain(int n)
{
    polymer_chain *chain;

    if(n < 2)
        return NULL;

    chain = malloc(sizeof(polymer_chain));

    chain->n = n;

    chain->r = malloc(sizeof(vector3) * n);
    chain->r_star = malloc(sizeof(vector3) * n);
    chain->f = malloc(sizeof(vector3) * n);
    chain->f_star = malloc(sizeof(vector3) * n);
    chain->w = malloc(sizeof(vector3) * n);

    return chain;
}

/*---------------------------------------------------------------------------*/
void free_polymer_chain(polymer_chain *chain)
{
    free(chain->r);
    free(chain->r_star);
    free(chain->f);
    free(chain->f_star);
    free(chain->w);
}

/*---------------------------------------------------------------------------*/
void initialize_polymer_chain(polymer_chain *chain,rng_state *rng,int n_b,
                              double *zeta_b)
{
    int i;

    if(n_b > 2)
        n_b = 2;

    chain->n_b = n_b;

    for(i = 0;i < n_b;i++)
    {
        if(zeta_b[i] > 0.0)
            chain->zeta_b_inv[i] = 1.0 / zeta_b[i];
        else
            chain->zeta_b_inv[i] = 1.0 / 0.0;
    }
    for(;i < 2;i++)
    {
        chain->zeta_b_inv[i] = 1.0;
    }

    chain->r[0][0] = 0;
    chain->r[0][1] = 0;
    chain->r[0][2] = 0;

    for(i = 1;i < chain->n;i++)
    {
        chain->r[i][0] = chain->r[i - 1][0] + nrand(rng) / sqrt(3);
        chain->r[i][1] = chain->r[i - 1][1] + nrand(rng) / sqrt(3);
        chain->r[i][2] = chain->r[i - 1][2] + nrand(rng) / sqrt(3);
    }
}

/*---------------------------------------------------------------------------*/
void calculate_stress_tensor(polymer_chain *chain,symmetric_tensor3 sigma)
{
    int i;
    vector3 *r;

    r = chain->r;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;
    for(i = 1;i < chain->n;i++)
    {
        sigma[0] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][0] - r[i - 1][0]);
        sigma[1] += 3 * (r[i][1] - r[i - 1][1]) * (r[i][1] - r[i - 1][1]);
        sigma[2] += 3 * (r[i][2] - r[i - 1][2]) * (r[i][2] - r[i - 1][2]);
        sigma[3] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][1] - r[i - 1][1]);
        sigma[4] += 3 * (r[i][0] - r[i - 1][0]) * (r[i][2] - r[i - 1][2]);
        sigma[5] += 3 * (r[i][1] - r[i - 1][1]) * (r[i][2] - r[i - 1][2]);
    }
}

/*---------------------------------------------------------------------------*/
void calculate_average_stress_tensor(polymer_chain **chain,int m,
                                     symmetric_tensor3 sigma)
{
    int j;
    symmetric_tensor3 sigma_1;

    sigma[0] = 0;
    sigma[1] = 0;
    sigma[2] = 0;
    sigma[3] = 0;
    sigma[4] = 0;
    sigma[5] = 0;
    for(j = 0;j < m;j++)
    {
        calculate_stress_tensor(chain[j],sigma_1);

        sigma[0] += sigma_1[0];
        sigma[1] += sigma_1[1];
        sigma[2] += sigma_1[2];
        sigma[3] += sigma_1[3];
        sigma[4] += sigma_1[4];
        sigma[5] += sigma_1[5];
    }

    sigma[0] /= m;
    sigma[1] /= m;
    sigma[2] /= m;
    sigma[3] /= m;
    sigma[4] /= m;
    sigma[5] /= m;
}

/*---------------------------------------------------------------------------*/
void calculate_center_of_mass(polymer_chain *chain,vector3 rc)
{
    int i;

    rc[0] = 0;
    rc[1] = 0;
    rc[2] = 0;
    for(i = 0;i < chain->n;i++)
    {
        rc[0] += chain->r[i][0];
        rc[1] += chain->r[i][1];
        rc[2] += chain->r[i][2];
    }
    rc[0] /= chain->n;
    rc[1] /= chain->n;
    rc[2] /= chain->n;
}

/*---------------------------------------------------------------------------*/
void calculate_inertia_tensor(polymer_chain *chain,symmetric_tensor3 g)
{
    int i;
    vector3 *r;
    vector3 rc;

    r = chain->r;

    calculate_center_of_mass(chain,rc);

    g[0] = 0;
    g[1] = 0;
    g[2] = 0;
    g[3] = 0;
    g[4] = 0;
    g[5] = 0;
    for(i = 0;i < chain->n;i++)
    {
        g[0] += (r[i][0] - rc[0]) * (r[i][0] - rc[0]);
        g[1] += (r[i][1] - rc[1]) * (r[i][1] - rc[1]);
        g[2] += (r[i][2] - rc[2]) * (r[i][2] - rc[2]);
        g[3] += (r[i][0] - rc[0]) * (r[i][1] - rc[1]);
        g[4] += (r[i][0] - rc[0]) * (r[i][2] - rc[2]);
        g[5] += (r[i][1] - rc[1]) * (r[i][2] - rc[2]);
    }
    g[0] /= chain->n;
    g[1] /= chain->n;
    g[2] /= chain->n;
    g[3] /= chain->n;
    g[4] /= chain->n;
    g[5] /= chain->n;
}

/*---------------------------------------------------------------------------*/
void calculate_average_inertia_tensor(polymer_chain **chain,int m,
                                      symmetric_tensor3 g)
{
    int j;
    symmetric_tensor3 g_1;

    g[0] = 0;
    g[1] = 0;
    g[2] = 0;
    g[3] = 0;
    g[4] = 0;
    g[5] = 0;
    for(j = 0;j < m;j++)
    {
        calculate_inertia_tensor(chain[j],g_1);

        g[0] += g_1[0];
        g[1] += g_1[1];
        g[2] += g_1[2];
        g[3] += g_1[3];
        g[4] += g_1[4];
        g[5] += g_1[5];
    }
    g[0] /= m;
    g[1] /= m;
    g[2] /= m;
    g[3] /= m;
    g[4] /= m;
    g[5] /= m;    
}

/*---------------------------------------------------------------------------*/
void calculate_maximum_distances(polymer_chain *chain,vector3 l_max)
{
    int i,j;

    l_max[0] = 0;
    l_max[1] = 0;
    l_max[2] = 0;
    for(i = 0;i < chain->n;i++)
    {
        for(j = i;j < chain->n;j++)
        {
            if(fabs(chain->r[j][0] - chain->r[i][0]) > l_max[0])
            {
                l_max[0] = fabs(chain->r[j][0] - chain->r[i][0]);
            }
            if(fabs(chain->r[j][1] - chain->r[i][1]) > l_max[1])
            {
                l_max[1] = fabs(chain->r[j][1] - chain->r[i][1]);
            }
            if(fabs(chain->r[j][2] - chain->r[i][2]) > l_max[2])
            {
                l_max[2] = fabs(chain->r[j][2] - chain->r[i][2]);
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
void calculate_average_maximum_distances(polymer_chain **chain,int m,
                                         vector3 l_max)
{
    int j;
    vector3 l_max_1;

    l_max[0] = 0;
    l_max[1] = 0;
    l_max[2] = 0;
    for(j = 0;j < m;j++)
    {
        calculate_maximum_distances(chain[j],l_max_1);

        l_max[0] += l_max_1[0];
        l_max[1] += l_max_1[1];
        l_max[2] += l_max_1[2];
    }
    l_max[0] /= m;
    l_max[1] /= m;
    l_max[2] /= m;
}

/*---------------------------------------------------------------------------*/
void calculate_end_to_end_vector(polymer_chain *chain,vector3 re)
{
    re[0] = chain->r[chain->n - 1][0] - chain->r[0][0];
    re[1] = chain->r[chain->n - 1][1] - chain->r[0][1];
    re[2] = chain->r[chain->n - 1][2] - chain->r[0][2];
}

/*---------------------------------------------------------------------------*/
void calculate_average_end_to_end_vector(polymer_chain **chain,int m,
                                         vector3 re)
{
    int j;
    vector3 re_1;

    re[0] = 0;
    re[1] = 0;
    re[2] = 0;
    for(j = 0;j < m;j++)
    {
        calculate_end_to_end_vector(chain[j],re_1);

        re[0] += re_1[0];
        re[1] += re_1[1];
        re[2] += re_1[2];
    }
    re[0] /= m;
    re[1] /= m;
    re[2] /= m;
}

/*---------------------------------------------------------------------------*/
