/*---------------------------------------------------------------------------*/

/*!  \file  integrator.c
     \brief  an integrator for the Langevin equation.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <math.h>
#include "integrator.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
static void clear_force(polymer_chain *chain);
static void generate_random_force(polymer_chain *chain,rng_state *rng);
static void calculate_spring_force(polymer_chain *chain,slip_springs *springs,
                                   int step);

/*---------------------------------------------------------------------------*/
#define SUFFIX linear
#undef BRANCHED_CHAIN
#include "integrator_kernel.c"
#undef SUFFIX

#define SUFFIX branched
#define BRANCHED_CHAIN 1
#include "integrator_kernel.c"
#undef BRANCHED_CHAIN
#undef SUFFIX

/*---------------------------------------------------------------------------*/
void integrate_langevin_equation(polymer_chain *chain,slip_springs *springs,
                                 rng_state *rng,double dt,tensor3 kappa)
{
    switch(chain->n_b)
    {
        case 0:
            integrate_langevin_equation_linear(chain,springs,rng,dt,kappa);
            break;

        case 1:
        case 2:
            integrate_langevin_equation_branched(chain,springs,rng,dt,kappa);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void clear_force(polymer_chain *chain)
{
    int i;

    for(i = 0;i < chain->n;i++)
    {
        chain->f[i][0] = 0;
        chain->f[i][1] = 0;
        chain->f[i][2] = 0;
        chain->f_star[i][0] = 0;
        chain->f_star[i][1] = 0;
        chain->f_star[i][2] = 0;
        chain->w[i][0] = 0;
        chain->w[i][1] = 0;
        chain->w[i][2] = 0;
    }
}

/*---------------------------------------------------------------------------*/
void generate_random_force(polymer_chain *chain,rng_state *rng)
{
    int i;

    for(i = 0;i < chain->n;i++)
    {
        chain->w[i][0] = nrand(rng);
        chain->w[i][1] = nrand(rng);
        chain->w[i][2] = nrand(rng);
    }
}

/*---------------------------------------------------------------------------*/
void calculate_spring_force(polymer_chain *chain,slip_springs *springs,
                            int step)
{
    int i,j;
    vector3 *r;
    vector3 *f;
    vector3 *a;

    switch(step)
    {
        case 1:
            r = chain->r;
            f = chain->f;
            a = springs->a;
            break;

        case 2:
            r = chain->r_star;
            f = chain->f_star;
            a = springs->a_star;
            break;

        default:
            return;
    }

    f[0][0] = 3 * (r[1][0] - r[0][0]);
    f[0][1] = 3 * (r[1][1] - r[0][1]);
    f[0][2] = 3 * (r[1][2] - r[0][2]);
    for(i = 1;i < (chain->n - 1);i++)
    {
        f[i][0] = 3 * (r[i + 1][0] + r[i - 1][0] - 2 * r[i][0]);
        f[i][1] = 3 * (r[i + 1][1] + r[i - 1][1] - 2 * r[i][1]);
        f[i][2] = 3 * (r[i + 1][2] + r[i - 1][2] - 2 * r[i][2]);
    }
    f[chain->n - 1][0] = 3 * (r[chain->n - 2][0] - r[chain->n - 1][0]);
    f[chain->n - 1][1] = 3 * (r[chain->n - 2][1] - r[chain->n - 1][1]);
    f[chain->n - 1][2] = 3 * (r[chain->n - 2][2] - r[chain->n - 1][2]);

    for(i = 0;i < springs->z;i++)
    {
        j = springs->s[i];

        f[j][0] += 3 / (springs->n_s) * (a[i][0] - r[j][0]);
        f[j][1] += 3 / (springs->n_s) * (a[i][1] - r[j][1]);
        f[j][2] += 3 / (springs->n_s) * (a[i][2] - r[j][2]);
    }
}

/*---------------------------------------------------------------------------*/
