/*---------------------------------------------------------------------------*/

/*!  \file  jump.c
     \brief  stochastic jump processes for slip-springs.  */

/*---------------------------------------------------------------------------*/

/* Copyright (C) 2009,2010 Takashi Uneyama  */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
   USA.  */

/*---------------------------------------------------------------------------*/
#include <math.h>
#include "jump.h"
#include "rng.h"

/*---------------------------------------------------------------------------*/
#define SQR(x) ((x) * (x))

/*---------------------------------------------------------------------------*/
static void swap_two_springs(slip_springs *springs,int i,int j);

/*---------------------------------------------------------------------------*/
#undef BRANCHED_CHAIN
#undef ENABLE_CR

#define SUFFIX linear
#include "jump_kernel.c"
#undef SUFFIX
#define ENABLE_CR 1
#define SUFFIX linear_cr
#include "jump_kernel.c"
#undef SUFFIX
#undef ENABLE_CR

#define BRANCHED_CHAIN 1
#define SUFFIX branched
#include "jump_kernel.c"
#undef BRANCHED_CHAIN
#undef SUFFIX
#define ENABLE_CR 1
#define SUFFIX branched_cr
#include "jump_kernel.c"
#undef SUFFIX
#undef ENABLE_CR
#undef BRANCHED_CHAIN

/*---------------------------------------------------------------------------*/
void move_slip_springs_by_jumps(polymer_chain *chain,slip_springs *springs,
                                rng_state *rng,double dt)
{
    double u;
    double w_plus,w_minus;
    double f,f_plus,f_minus;
    double zeta_s_inv;
    int i,j;

    zeta_s_inv = springs->zeta_s_inv;

    for(i = 0;i < springs->z;i++)
    {
        j = springs->s[i];
        f = 3.0 / (2.0 * springs->n_s)
            * (SQR(chain->r[j][0] - springs->a[i][0])
               + SQR(chain->r[j][1] - springs->a[i][1])
               + SQR(chain->r[j][2] - springs->a[i][2]));

        if(j < (chain->n - 1))
        {
            f_plus = 3.0 / (2.0 * springs->n_s)
                * (SQR(chain->r[j + 1][0] - springs->a[i][0])
                   + SQR(chain->r[j + 1][1] - springs->a[i][1])
                   + SQR(chain->r[j + 1][2] - springs->a[i][2]));

            w_plus = (dt * zeta_s_inv)
                * (1.0 - tanh(0.5 * (f_plus - f)));

            if(w_plus > 0.5)
            {
                w_plus = 0.5;
            }
        }
        else
        {
            w_plus = 0.0;
        }

        if(j > 0)
        {
            f_minus = 3.0 / (2.0 * springs->n_s)
                * (SQR(chain->r[j - 1][0] - springs->a[i][0])
                   + SQR(chain->r[j - 1][1] - springs->a[i][1])
                   + SQR(chain->r[j - 1][2] - springs->a[i][2]));

            w_minus = (dt * zeta_s_inv)
                * (1.0 - tanh(0.5 * (f_minus - f)));

            if(w_minus > 0.5)
            {
                w_minus = 0.5;
            }
        }
        else
        {
            w_minus = 0.0;
        }


        u = urand(rng);
        if(u < w_plus)
        {
            springs->s[i]++;
        }
        else if(u < (w_plus + w_minus))
        {
            springs->s[i]--;
        }
    }
}

/*---------------------------------------------------------------------------*/
void reconstruct_slip_springs(polymer_chain *chain,slip_springs *springs,
                              rng_state *rng,double dt)
{
    switch(chain->n_b)
    {
        case 0:
            if(springs->tau_s_inv == 0.0)
                reconstruct_slip_springs_linear(chain,springs,rng,dt);
            else
                reconstruct_slip_springs_linear_cr(chain,springs,rng,dt);
            break;

        case 1:
        case 2:
            if(springs->tau_s_inv == 0.0)
                reconstruct_slip_springs_branched(chain,springs,rng,dt);
            else
                reconstruct_slip_springs_branched_cr(chain,springs,rng,dt);
            break;

        default:
            break;
    }
}

/*---------------------------------------------------------------------------*/
void swap_two_springs(slip_springs *springs,int i,int j)
{
    int s_save;
    vector3 a_save;

    s_save = springs->s[i];
    a_save[0] = springs->a[i][0];
    a_save[1] = springs->a[i][1];
    a_save[2] = springs->a[i][2];

    springs->s[i] = springs->s[j];
    springs->a[i][0] = springs->a[j][0];
    springs->a[i][1] = springs->a[j][1];
    springs->a[i][2] = springs->a[j][2];

    springs->s[j] = s_save;
    springs->a[j][0] = a_save[0];
    springs->a[j][1] = a_save[1];
    springs->a[j][2] = a_save[2];
}

/*---------------------------------------------------------------------------*/
